/*
* Copyright (C) 2011 Andrew Reid and the modelGUI Project <http://mgui.wikidot.com>
* 
* This file is part of modelGUI[neuro] (mgui-neuro).
* 
* modelGUI[neuro] is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* modelGUI[neuro] is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with modelGUI[neuro]. If not, see <http://www.gnu.org/licenses/>.
*/

package mgui.io.foreign.minc;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;

import mgui.datasources.DataTypes;
import mgui.image.util.WindowedColourModel;
import mgui.interfaces.InterfaceSession;
import mgui.interfaces.ProgressUpdater;
import mgui.interfaces.logs.LoggingType;
import mgui.interfaces.shapes.Volume3DInt;
import mgui.interfaces.shapes.volume.GridVertexDataColumn;
import mgui.interfaces.shapes.volume.VolumeMetadata;
import mgui.io.domestic.shapes.ShapeIOException;
import mgui.io.domestic.shapes.VolumeFileLoader;
import mgui.io.domestic.shapes.VolumeInputOptions;
import mgui.numbers.MguiNumber;
import mgui.stats.StatFunctions;


/************************************
 * Loads Medical Imaging NetCDF (MINC) 2.0 format files, which are based on the HDF5 library.
 * Note that this loader requires the HDF5 Java library (see <a href="http://www.hdfgroup.org/hdf-java-html/hdfview/">
 * http://www.hdfgroup.org/hdf-java-html/hdfview/</a>). This requires a JNI binding, and is therefore not OS
 * independent. Since this violates the platform independence of ModelGUI (based on a "pure Java" implementation),
 * inclusion of this loader will be made available as a non-core functionality of the Project.
 * 
 * 
 * @author Andrew Reid
 * @version 1.0
 * @since 1.0
 *
 */

public class MincVolumeLoader extends VolumeFileLoader {

	public MincVolumeLoader(){
		
	}
	
	public MincVolumeLoader(File file){
		this.dataFile = file;
	}

	public VolumeMetadata getVolumeMetadata() throws IOException, FileNotFoundException{
		if (dataFile == null) 
			throw new IOException("No data file specified.");
		if (!dataFile.exists() || !dataFile.isFile())
			throw new FileNotFoundException("File '" + dataFile.getAbsolutePath() + "' not found.");
		MincVolumeMetadata metadata =  new MincVolumeMetadata(dataFile);
		if (metadata.isInit())
			return new MincVolumeMetadata(dataFile);
		return null;
	}

	@Override
	protected boolean setVolume3DBlocking(Volume3DInt volume, String column,
											int v, VolumeInputOptions options, ProgressUpdater progress)
											throws ShapeIOException {
		
		try{
			MincVolumeMetadata metadata = new MincVolumeMetadata(dataFile);
			metadata.options = (MincVolumeInputOptions)options;
			
			ArrayList<MguiNumber> values = metadata.readVolume(v, 
															   DataTypes.getFromDataBufferType(options.transfer_type), 
															   progress);
			
			flipX = options.flip_x;
			flipY = options.flip_y;
			flipZ = options.flip_z;
			
			volume.addVertexData(column, values, options.colour_map);
			volume.hasAlpha(options.has_alpha);
			volume.setCurrentColumn(column, false);
			
			GridVertexDataColumn v_column = (GridVertexDataColumn)volume.getVertexDataColumn(column);
			double[] stats = StatFunctions.getBasicNormalStats(values);
			double min = stats[4];
			double max = stats[5];
			if (v_column == null){
				return false;
			  	}
			WindowedColourModel cm = v_column.getColourModel();
			  
			//intercept and scale
			cm.setIntercept(min);
			  if (max > min)
				  cm.setScale(1.0 / (max - min));
			  else
				  cm.setScale(1.0);
			 cm.setWindowMid(0.5);
			 cm.setWindowWidth(1.0);
			 v_column.setDataMin(min, false);
			 v_column.setDataMax(max, false);
			 v_column.setColourLimits(min, max, false);
			  
			InterfaceSession.log("Scale: " + cm.getScale() + " Intercept: " + cm.getIntercept(), LoggingType.Debug);
			InterfaceSession.log("Min: " + min + " Max: " + max, LoggingType.Debug);
			  
			return true;
			
		}catch (IOException e){
			//e.printStackTrace();
			InterfaceSession.handleException(e);
			return false;
			}
		
	}	
	

}