/*
* Copyright (C) 2011 Andrew Reid and the modelGUI Project <http://mgui.wikidot.com>
* 
* This file is part of modelGUI[neuro] (mgui-neuro).
* 
* modelGUI[neuro] is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* modelGUI[neuro] is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with modelGUI[neuro]. If not, see <http://www.gnu.org/licenses/>.
*/

package mgui.io.foreign.minc;

import java.awt.image.DataBuffer;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import mgui.datasources.DataType;
import mgui.datasources.DataTypes;
import mgui.geometry.Box3D;
import mgui.interfaces.InterfaceSession;
import mgui.interfaces.ProgressUpdater;
import mgui.interfaces.logs.LoggingType;
import mgui.interfaces.shapes.Volume3DInt;
import mgui.interfaces.shapes.volume.VolumeMetadata;
import mgui.numbers.MguiDouble;
import mgui.numbers.MguiFloat;
import mgui.numbers.MguiInteger;
import mgui.numbers.MguiNumber;
import mgui.numbers.MguiShort;
import ncsa.hdf.object.Attribute;
import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Datatype;
import ncsa.hdf.object.FileFormat;
import ncsa.hdf.object.Group;
import ncsa.hdf.object.h4.H4File;
import ncsa.hdf.object.h5.H5File;
import Jama.Matrix;


/***************************************************************
 * Metadata for a Medical Imaging NetCDF (MINC) 2.0 format file.
 * 
 * @author typically
 * @version 1.0
 * @since 1.0
 *
 */
public class MincVolumeMetadata implements VolumeMetadata {

	protected Matrix cosines;		// Cosines determining world coordinate system
	protected Matrix start; 		// Start point, in real coordinates
	protected float[] step; 		// Step size, in real coordinates
	protected int[] dims, in_dims, out_dims; 	// Voxel dimensions (s,t,r) and input ordered
	protected int[] dim_order; 		// Dimension ordering
	protected int tdim;				// Time dimension 
	int datatype; 					// Data type of stored values
	
	// For output
	float[] geom_dims; 				// Geometric dimensions
	Point3f origin;					// Origin, in world coordinates
	
	// HDF5 objects
	FileFormat hdfile; 				// HDF4 or HDF5 format file	
	Group root; 					// Root of HDF5 data structure
	Group images_group; 			// Group containing the image data
	Dataset image_ds; 				// Image dataset, for malformed files
	
	private boolean is_init = false;
	
	MincVolumeInputOptions options = new MincVolumeInputOptions();
	
	/******************************************
	 * Constructs a new instance with the given input file.
	 * 
	 * @param file
	 */
	public MincVolumeMetadata(File file){
		
		init(file);
		
	}
	
	protected void init(File file){
		init(file, true);
	}
	
	private int format = 5;
	
	protected boolean init(File file, boolean close){
		is_init = false;
		if (!file.exists()){
			InterfaceSession.log("MincVolumeMetadata.init: File not found: '" + file.getAbsolutePath() + "'.", 
					 LoggingType.Errors);
			return false;
			}
		try{
			hdfile = new H5File(file.getAbsolutePath());
			hdfile.open();
			format = 5;
			InterfaceSession.log("MINC-2.0 format detected.", LoggingType.Debug);
		}catch (Exception ex){
			try{
				hdfile = new H4File(file.getAbsolutePath());
				hdfile.open();
				format = 4;
				InterfaceSession.log("MINC-1.0 format detected.", LoggingType.Debug);
			}catch (Exception ex2){
				InterfaceSession.log("Error opening MINC file '" + file.getAbsolutePath() + "'", 
									 LoggingType.Errors);
				return false;
				}
			}
		
		try{
			
			root = (Group)(( DefaultMutableTreeNode)hdfile.getRootNode()).getUserObject();
			format = 5;
			
			// Search the tree, populate as we go
			if (!queryHDF5Tree(root) || !found_dimensions || !found_images){
				format = 4;
				// This is likely a MINC-1.0 format, load accordingly
				if (!queryHDF4Tree(root) || !found_dimensions || !found_images){
					// This is a malformed tree, try to query as a flat set of datasets
					if (!queryMalformedTree(root)){
						format = 0;
						InterfaceSession.log("MincVolumeMetadata.init: Invalid MINC header: not all header " +
											 "information was present.", 
											 LoggingType.Errors);
						hdfile.close();
						return false;
						}
					}
				}
			
			is_init = true;
			if (close) hdfile.close();
			return true;
			
		}catch (Exception ex){
			InterfaceSession.handleException(ex);
			return false;
			}
	}
	
	public boolean isInit(){
		return is_init;
	}
	
	public boolean found_dimensions = false;
	public boolean found_images = false;
	
	// Query the HDF5 tree for the information we want
	protected boolean queryHDF5Tree(Group group){
		
		String path = group.getFullName();
		
		if (path.equals("/minc-2.0/dimensions")){
			found_dimensions = queryDimensions(group);
			return found_dimensions;
			}
		
		if (path.equals("/minc-2.0/image")){
			found_images = queryImages(group);
			return found_images;
			}
		
		List<?> children = group.getMemberList();
		for (int i = 0; i < children.size(); i++){
			if (children.get(i) instanceof Group){
				if (!queryHDF5Tree((Group)children.get(i))) return false;
			}else{
				// This is a data set, do nothing
				
				}
			}
		
		// If we get here there have been no problems
		return (!(cosines == null || dims == null || dim_order == null || start == null || step == null));
	}
	
	protected boolean queryMalformedTree(Group group){
		
		List<?> children = group.getMemberList();
		for (int i = 0; i < children.size(); i++){
			if (children.get(i) instanceof Dataset){
				if (querySpaceDataMinc1((Dataset)children.get(i))){
					found_dimensions |= true;
					}
				if (queryImageDataMinc1((Dataset)children.get(i))){
					found_images |= true;
					}
				}
			}
		
		// If we get here there have been no problems
		return (!(cosines == null || dims == null || dim_order == null || start == null || step == null));
		
	}
	
	// Query the Dimensions group 
	private boolean queryDimensions(Group group){
		
		List<?> children = group.getMemberList();
		cosines = null;
		dims = null;
		start = null;
		step = null;
		
		try{
			for (int i = 0; i < children.size(); i++){
				if (children.get(i) instanceof Dataset){
					Dataset ds = (Dataset)children.get(i);
					int dim = -1;
					if (ds.getName().equals("xspace")) dim = 0;
					if (ds.getName().equals("yspace")) dim = 1;
					if (ds.getName().equals("zspace")) dim = 2;
					
					List<?> attributes = ds.getMetadata();
					for (int j = 0; j < attributes.size(); j++){
						Attribute attr = (Attribute)attributes.get(j);
						if (attr.getName().equals("direction_cosines")){
							double[] cos = (double[])attr.getValue();
							if (cosines == null) cosines = new Matrix(3,3);
							for (int k = 0; k < 3; k++)
								cosines.set(dim, k, cos[k]);
							}
						if (attr.getName().equals("length")){
							if (dims == null) dims = new int[3];
							dims[dim] = ((int[])attr.getValue())[0];
							}
						if (attr.getName().equals("start")){
							if (start == null) start = new Matrix(3,1);
							start.set(dim, 0, ((double[])attr.getValue())[0]);
							}
						if (attr.getName().equals("step")){
							if (step == null) step = new float[3];
							step[dim] = (float)((double[])attr.getValue())[0];
							}
						}
					}
				}
			
		}catch (Exception ex){
			InterfaceSession.handleException(ex);
			return false;
			}
		
		return (!(cosines == null || dims == null || start == null || step == null));
	}
	
	private boolean querySpaceDataMinc1(Dataset ds){
		
		int dim = -1;
		if (ds.getName().equals("xspace")) dim = 0;
		if (ds.getName().equals("yspace")) dim = 1;
		if (ds.getName().equals("zspace")) dim = 2;
		if (dim < 0) return false;
		try{
			List<?> attributes = ds.getMetadata();
			for (int j = 0; j < attributes.size(); j++){
				Attribute attr = (Attribute)attributes.get(j);
				if (attr.getName().equals("start")){
					if (start == null) start = new Matrix(3,1);
					start.set(dim, 0, ((double[])attr.getValue())[0]);
					}
				if (attr.getName().equals("step")){
					if (step == null) step = new float[3];
					step[dim] = (float)((double[])attr.getValue())[0];
					}
				}
		}catch (Exception ex){
			InterfaceSession.handleException(ex);
			return false;
			}
		
		return true;
	}
	
	// Query the Images group 
	private boolean queryImages(Group group){
		
		images_group = group;
		List<?> children = images_group.getMemberList();
		if (children.size() < 1) return false;
		
		Group image0 = (Group)children.get(0);
		try{
			children = image0.getMemberList();
		for (int j = 0; j < children.size(); j++){
			Dataset ds = (Dataset)children.get(j);
			if (ds.getName().equals("image")){
				long[] sel = ds.getSelectedDims();
//				sel[0]=1;sel[1]=1;sel[2]=1;
//				ds.read();
				this.datatype = getDataTypeForHDF(ds.getDatatype());
				List<?> attributes = ds.getMetadata();
				for (int k = 0; k < attributes.size(); k++){
					Attribute attr = (Attribute)attributes.get(k);
					if (attr.getName().equals("dimorder")){
						String[] dm = (String[])attr.getValue();
						dm = (dm[0]).split(",");
						if (dm != null){
							if (dim_order == null)
								dim_order = new int[3];
							for (int l = 0; l < 3; l++){
								if (dm[l].equals("xspace")) dim_order[l] = 0;
								if (dm[l].equals("yspace")) dim_order[l] = 1;
								if (dm[l].equals("zspace")) dim_order[l] = 2;
								}
							}
						}
					}
				return true;
				}
			}
		}catch (Exception ex){
			InterfaceSession.handleException(ex);
			}
		
		return false;
		
	}
	
	private boolean queryImageDataMinc1(Dataset ds){
		if (!ds.getName().equals("image")) return false;
		
		try{
			
			ds.read();
			this.datatype = getDataTypeForHDF(ds.getDatatype());
			
			String[] dm = ds.getDimNames();
			if (dm != null){
				if (dim_order == null)
					dim_order = new int[3];
				for (int k = 0; k < 3; k++){
					if (dm[k].equals("xspace")) dim_order[k] = 0;
					if (dm[k].equals("yspace")) dim_order[k] = 1;
					if (dm[k].equals("zspace")) dim_order[k] = 2;
					}
				}
			
			long[] dims = ds.getDims();
			if (this.dims == null){
				if (dim_order != null)
					this.dims = new int[]{(int)dims[dim_order[0]],(int)dims[dim_order[1]],(int)dims[dim_order[2]]};
				else
					this.dims = new int[]{(int)dims[0],(int)dims[1],(int)dims[2]};
			}
			cosines = new Matrix(3,3);
			for (int i = 0; i < 3; i++)
				cosines.set(i, i, 1);
			
			image_ds = ds;
			return true;
		}catch (Exception ex){
			InterfaceSession.handleException(ex);
			}
		return false;
	}
	
	protected boolean queryHDF4Tree(Group group){
		
		cosines = null;
		dim_order = null;
		dims = null;
		start = null;
		step = null;
		images_group = null;
		datatype = -1;
		
		queryHDF4TreeRecursive(group);
	
		// Is everything present?
		if (image_ds == null) return false;
		found_images = true;
		
		if (cosines == null || dims == null || start == null || step == null) return false;
		found_dimensions = true;
		
		return true;
		
	}
	
	private void queryHDF4TreeRecursive(Group group){
		
		List<?> children = group.getMemberList();
		
		try{
			for (int i = 0; i < children.size(); i++){
				Object child = children.get(i);
				if (child instanceof Group){
					queryHDF4TreeRecursive((Group)child);
				}else{
					Dataset ds = (Dataset)child;
					if (ds.getName().endsWith("space")){
						int dim = -1;
						if (ds.getName().equals("xspace")) dim = 0;
						if (ds.getName().equals("yspace")) dim = 1;
						if (ds.getName().equals("zspace")) dim = 2;
						
						List<?> attributes = ds.getMetadata();
						for (int j = 0; j < attributes.size(); j++){
							Attribute attr = (Attribute)attributes.get(j);
							if (attr.getName().equals("direction_cosines")){
								double[] cos = (double[])attr.getValue();
								if (cosines == null) cosines = new Matrix(3,3);
								for (int k = 0; k < 3; k++)
									cosines.set(dim, k, cos[k]);
								}
							if (attr.getName().equals("start")){
								if (start == null) start = new Matrix(3,1);
								start.set(dim, 0, ((double[])attr.getValue())[0]);
								}
							if (attr.getName().equals("step")){
								if (step == null) step = new float[3];
								step[dim] = (float)((double[])attr.getValue())[0];
								}
							}
						} // Spatial information
					
					if (ds.getName().endsWith("image")){
						// Single image
						dim_order = new int[]{2,1,0}; 		// Default ordering
						image_ds = ds;
						ds.read(); 			// Has to be called before anything sensible happens.. bug?
						long[] ldims = ds.getDims();
						dims = new int[]{(int)ldims[0],(int)ldims[1],(int)ldims[2]};
						found_images = true;
						datatype = getDataTypeForHDF(ds.getDatatype());
						String[] dm = ds.getDimNames();
						if (dm != null){
							for (int k = 0; k < 3; k++){
								if (dm[k].equals("xspace")) dim_order[k] = 0;
								if (dm[k].equals("yspace")) dim_order[k] = 1;
								if (dm[k].equals("zspace")) dim_order[k] = 2;
								}
						}else{
							List<?> attributes = image_ds.getMetadata();
							for (int j = 0; j < attributes.size(); j++){
								Attribute attr = (Attribute)attributes.get(j);
								if (attr.getName().equals("dimorder")){
									dim_order = new int[3];
									String val = ((String[])attr.getValue())[0];
									dm = val.split(",");
									for (int k = 0; k < 3; k++){
										if (dm[k].equals("xspace")) dim_order[k] = 0;
										if (dm[k].equals("yspace")) dim_order[k] = 1;
										if (dm[k].equals("zspace")) dim_order[k] = 2;
										}
									}
								}
							}
						dims = new int[]{dims[dim_order[0]],dims[dim_order[1]],dims[dim_order[2]]};
						}
					
					}
				
				}
			}catch (Exception ex){
				InterfaceSession.handleException(ex);
				return;
				}
		
	}
	
	protected int getDataTypeForHDF(Datatype type) throws Exception{
		
		int size = type.getDatatypeSize();
		
		switch (type.getDatatypeClass()){
		
			case Datatype.CLASS_CHAR:
				switch (size){
				case 2:
					return DataBuffer.TYPE_SHORT;
				case 1:
					return DataBuffer.TYPE_BYTE;
				}
		
			case Datatype.CLASS_INTEGER:
				switch (size){
					case 4:
						if (type.getDatatypeSign() == Datatype.SIGN_NONE)
							return DataBuffer.TYPE_UNDEFINED;
						else
							return DataBuffer.TYPE_INT;
					case 2:
						if (type.getDatatypeSign() == Datatype.SIGN_NONE)
							return DataBuffer.TYPE_USHORT;
						else
							return DataBuffer.TYPE_SHORT;
					case 1:
						return DataBuffer.TYPE_BYTE;
						}
				
				break;
				
			case Datatype.CLASS_FLOAT:
			case Datatype.CLASS_NO_CLASS:
				switch (size){
					case 4:
						return DataBuffer.TYPE_FLOAT;
					case 8:
					default:
						return DataBuffer.TYPE_DOUBLE;
					}
				
			}
		
		throw new Exception("Bad data type for MINC image data: " + type.getDatatypeDescription());
			
	}
	
	@Override
	public void setFromMetadata(VolumeMetadata metadata){
		
		this.setOrigin(metadata.getOrigin());
		this.setDataType(metadata.getDataType());
		this.setDataDims(metadata.getDataDims());
		this.setGeomDims(metadata.getGeomDims());
		this.setAxes(metadata.getAxes());
		
	}
	
	@Override
	public int[] getDataDims() {
		return dims;
//		if (dims == null || dim_order == null) return null;
//		return new int[]{dims[dim_order[0]],dims[dim_order[1]],dims[dim_order[2]]};
	}
	
	@Override
	public float[] getGeomDims() {
		float[] geom = new float[3];
		geom[0] = (float)dims[0] * step[0];
		geom[1] = (float)dims[1] * step[1];
		geom[2] = (float)dims[2] * step[2];
		return geom;
	}
	
	@Override
	public Vector3f[] getAxes() {
		Vector3f[] axes = new Vector3f[3];
		for (int i = 0; i < 3; i++)
			axes[i] = new Vector3f((float)cosines.get(i, 0), 
								   (float)cosines.get(i, 1), 
								   (float)cosines.get(i, 2));
		return axes;
	}
	
	@Override
	public int getVoxelDim() {
		return tdim;
	}
	
	@Override
	public Point3f getOrigin() {
		Matrix O = new Matrix(3,1);
		
		O = start.transpose().times(cosines);
		Point3f p = new Point3f((float)O.get(0, 0), 
								(float)O.get(0, 1), 
								(float)O.get(0, 2));
		return p;
	}
	
	@Override
	public int getDataType() {
		return datatype;
	}
	
	@Override
	public Box3D getBounds() {
		Vector3f[] axes = getAxes();
		Point3f origin = getOrigin();
		// Set bounds origin back half a voxel
		axes[0].scale(step[0] * 0.5f);
		axes[1].scale(step[1] * 0.5f);
		axes[2].scale(step[2] * 0.5f);
		origin.sub(axes[0]);
		origin.sub(axes[1]);
		origin.sub(axes[2]);
		
		float[] geom = getGeomDims();
		axes = getAxes();
		// Scale axes by geometric dimensions
		axes[0].scale(geom[0]);
		axes[1].scale(geom[1]);
		axes[2].scale(geom[2]);
		
		Box3D bounds = new Box3D(origin, axes[0], axes[1], axes[2]);
		return bounds;
	}
	@Override
	public void setDataDims(int[] dims) {
		this.dims = dims;
	}
	@Override
	public void setGeomDims(float[] dims) {
		this.geom_dims = dims;
	}
	@Override
	public void setAxes(Vector3f[] axes) {
		cosines = new Matrix(3,3);
		for (int i = 0; i < 3; i++){
			cosines.set(i, 0, axes[i].getX());
			cosines.set(i, 1, axes[i].getY());
			cosines.set(i, 2, axes[i].getZ());
			}
	}
	@Override
	public void setVoxelDim(int t) {
		tdim = t;
	}
	@Override
	public void setOrigin(Point3f origin) {
		this.origin = origin;
	}
	@Override
	public void setDataType(int type) {
		this.datatype = type;
	}
	@Override
	public void setFromVolume(Volume3DInt volume) {
		// TODO Auto-generated method stub
		
	}
	@Override
	public void setFromVolume(Volume3DInt volume, String column) {
		// TODO Auto-generated method stub
		
	}
	@Override
	public ArrayList<MguiNumber> readVolume(int t, DataType data_type, ProgressUpdater progress) throws IOException {
		
		if (!isInit()) throw new IOException("Cannot initiate MINC file from header.");
		
		if (!found_images || (images_group == null && image_ds == null)) throw new IOException("No image group found in MINC file.");
		
		//if (images_group == null) return readMinc1(t, data_type, progress);
		
		try{
			init(hdfile.getAbsoluteFile(),false);
			
			double[] image_min = null, image_max = null;
			double[] range = null;
			
			boolean is_minc1 = images_group == null;
			
			if (!is_minc1){
				List<?> volumes = images_group.getMemberList();
				Group volume_t = null;
				
				for (int i = 0; i < volumes.size(); i++){
					Group volume_i = (Group)volumes.get(i);
					try{
						int idx = Integer.valueOf(volume_i.getName());
						if (idx == t) 
							volume_t = volume_i; 
					}catch (NumberFormatException ex){
						}
					}
				List<?> children = volume_t.getMemberList();
				for (int i = 0; i < children.size(); i++){
					Dataset ds = (Dataset)children.get(i);
					if (ds.getName().equals("image")){
						image_ds = ds;
						}
					if (options.scale_slicewise && ds.getName().equals("image-min")){
						switch(datatype){
							case DataBuffer.TYPE_FLOAT:
							case DataBuffer.TYPE_DOUBLE:
								image_min = (double[])ds.getData();
								break;
							default:
								int[] vals = (int[])ds.getData();
								image_min = new double[vals.length];
								for (int j = 0; j < vals.length; j++)
									image_min[j] = (double)vals[j];
							}
						}
					
					if (options.scale_slicewise && ds.getName().equals("image-max")){
						switch(datatype){
							case DataBuffer.TYPE_FLOAT:
							case DataBuffer.TYPE_DOUBLE:
								image_max = (double[])ds.getData();
								break;
							default:
								int[] vals = (int[])ds.getData();
								image_max = new double[vals.length];
								for (int j = 0; j < vals.length; j++)
									image_max[j] = (double)vals[j];
							}
						}
					}
				}
			
			if (image_ds == null)
				throw new IOException("No volume found at time point " + t + ".");
			
			
			Datatype hdf_datatype = null;
			int mgui_datatype = -1;
			
			image_ds.read();
			hdf_datatype = image_ds.getDatatype();
			
			mgui_datatype = this.getDataTypeForHDF(hdf_datatype);
			List<?> attributes = image_ds.getMetadata();
			for (int j = 0; j < attributes.size(); j++){
				Attribute attr = (Attribute)attributes.get(j);
				
				if (attr.getName().equals("valid_range")){
					Datatype type = attr.getType();
					Object rr = null;
					switch (type.getDatatypeClass()){
						case Datatype.CLASS_INTEGER:
							rr = (int[])attr.getValue();
							range = new double[]{((int[])rr)[0], ((int[])rr)[1]};
							break;
						case Datatype.CLASS_FLOAT:
							if (type.getDatatypeSize() < 8){
								rr = (float[])attr.getValue();
								range = new double[]{((float[])rr)[0], ((float[])rr)[1]};
							}else{
								range = (double[])attr.getValue();
								}
							break;
						}
					}
				}
		
			if (dim_order == null){
				InterfaceSession.log("MincVolumeMetadata: Dimension ordering not specified in MINC file; " +
									 "assuming x,y,z..", LoggingType.Warnings);
				dim_order = new int[]{0,1,2}; 		// Default ordering
				}
			
			in_dims = new int[]{dims[dim_order[0]],dims[dim_order[1]],dims[dim_order[2]]};
			int n_voxels = dims[0]*dims[1]*dims[2];
			
			int data_type_out = DataTypes.getDataBufferType(data_type);
			
			double v_in = Double.NaN;
			
			long[] start_dims = image_ds.getStartDims();
			long[] selected_dims = image_ds.getSelectedDims();
			long[] stride = image_ds.getStride();
			int[] selected_idx = image_ds.getSelectedIndex();
			selected_idx[0] = dim_order[1];
			selected_idx[1] = dim_order[2];
			selected_idx[2] = dim_order[0];
			selected_dims[0] = 1;
			selected_dims[1] = in_dims[1];
			selected_dims[2] = in_dims[2];
			stride[0] = 1;
			stride[1] = 1;
			stride[2] = 1;
			Object data;
			
			if (progress != null){
				progress.setMinimum(0);
				progress.setMaximum(in_dims[0]);
				}
			
			if (progress != null){
				progress.setMinimum(0);
				progress.setMaximum(in_dims[0] * 2);
				}
			ArrayList<MguiNumber> values = new ArrayList<MguiNumber>(n_voxels);
			int chunk = in_dims[1] * in_dims[2];
			int j = 0, itr = 0;
			// Fill data array
			for (int i = 0; i < n_voxels; i++){
				switch (data_type_out){
					case DataBuffer.TYPE_FLOAT:
						values.add(new MguiFloat(v_in));
						break;
					case DataBuffer.TYPE_DOUBLE:
						values.add(new MguiDouble(v_in));
						break;
					case DataBuffer.TYPE_UNDEFINED: // Unsigned int
						values.add(new MguiFloat(v_in));
						break;
					case DataBuffer.TYPE_INT:
						values.add(new MguiInteger(v_in));
						break;
					case DataBuffer.TYPE_SHORT:
						values.add(new MguiShort(v_in));
						break;
					case DataBuffer.TYPE_USHORT:
						values.add(new MguiInteger(v_in));
						break;
					case DataBuffer.TYPE_BYTE:
						values.add(new MguiShort(v_in));
					}
				if (progress != null && ++j == chunk){
					j = 0;
					progress.update(itr++);
					}
				}
			
			for (int i = 0; i < in_dims[0]; i++){
				// Each i is a new image slice, must scale it independently
				double min, max;
				if (options.scale_slicewise && image_min != null){
					min = image_min[i];
					max = image_max[i];
				}else{
					min = range[0];
					max = range[1];
					}
				start_dims[0] = i;
				image_ds.clearData();
				data = image_ds.getData();
				min += 0;
				int abs_idx=0;
				
				for (j = 0; j < in_dims[1]; j++){
					for (int k = 0; k < in_dims[2]; k++){
						int idx = j*in_dims[2] + k;
						switch(mgui_datatype){
							case DataBuffer.TYPE_FLOAT:
								v_in = ((float[])data)[idx];
								break;
							case DataBuffer.TYPE_DOUBLE:
								v_in = ((double[])data)[idx];
								break;
							case DataBuffer.TYPE_UNDEFINED:		// Signifies an unsigned int here
								v_in = ((int[])data)[idx] & 0xffffffffL; 		// Translate to float
								break;
							case DataBuffer.TYPE_INT:
								v_in = ((int[])data)[idx];
								break;
							case DataBuffer.TYPE_USHORT:
								v_in = ((short[])data)[idx] & 0xffff;			// Translate to int
								break;
							case DataBuffer.TYPE_SHORT:
								v_in = ((short[])data)[idx];
								break;
							case DataBuffer.TYPE_BYTE:
								v_in = ((byte[])data)[idx];
								break;
							default:
								v_in = ((double[])data)[idx]; // Will throw an exception
							}
						// Scale value
						v_in = scaleData(v_in, range[0], range[1], min, max);
//						if (is_minc1)
//							abs_idx = getAbsoluteIndex1(dim_order,new int[]{i,j,k});
//						else
						
						abs_idx = getAbsoluteIndex(dim_order,new int[]{i,j,k});
						values.get(abs_idx).setValue(v_in);
						
						}
					}
				if (progress != null){
					progress.update(itr++);
					}
				}
			
			return values;
		
		}catch (Exception ex){
			throw new IOException(ex);
			}
		
	}
	
	private int getAbsoluteIndex1(int[] dimorder, int[] idx){
		
		return idx[2] * dims[1] * dims[0] + idx[1] * dims[0] + idx[0];
		
	}
	
	// Put the dimorder in x,y,z order and return the absolute index
	private int getAbsoluteIndex(int[] dimorder, int[] idx){
		
		int x_idx = -1, y_idx = -1, z_idx = -1;
		for (int i = 0; i < 3; i++){
			if (dimorder[i] == 0) x_idx = idx[i];
			if (dimorder[i] == 1) y_idx = idx[i];
			if (dimorder[i] == 2) z_idx = idx[i];
			}
			
		return z_idx * dims[1] * dims[0] + y_idx * dims[0] + x_idx;
		
	}
	
	private double scaleData(double val, double min_from, double max_from, double min_to, double max_to){
		
		val = (val - min_from) / (max_from - min_from);
		val = val * (max_to - min_to) + min_to;
		return val;
		
	}
	
	
	protected ArrayList<MguiNumber> readMinc1(int t, DataType data_type, ProgressUpdater progress) throws IOException {
		
		try{
			init(hdfile.getAbsoluteFile(),false);
			
			double[] image_min = null, image_max = null;
			double[] range = null;
			
			if (image_ds == null)
				throw new IOException("No volume found at time point " + t + ".");
			
			Datatype hdf_datatype = null;
			int mgui_datatype = -1;
			
			image_ds.read();
			hdf_datatype = image_ds.getDatatype();
			
			mgui_datatype = this.getDataTypeForHDF(hdf_datatype);
			List<?> attributes = image_ds.getMetadata();
			for (int j = 0; j < attributes.size(); j++){
				Attribute attr = (Attribute)attributes.get(j);
				
				if (attr.getName().equals("valid_range")){
					Datatype type = attr.getType();
					Object rr = null;
					switch (type.getDatatypeClass()){
						case Datatype.CLASS_INTEGER:
							rr = (int[])attr.getValue();
							range = new double[]{((int[])rr)[0], ((int[])rr)[1]};
							break;
						case Datatype.CLASS_FLOAT:
							if (type.getDatatypeSize() < 8){
								rr = (float[])attr.getValue();
								range = new double[]{((float[])rr)[0], ((float[])rr)[1]};
							}else{
								range = (double[])attr.getValue();
								}
							break;
						}
					}
				}
		
			if (dim_order == null){
				InterfaceSession.log("MincVolumeMetadata: Dimension ordering not specified in MINC file; " +
									 "assuming x,y,z..", LoggingType.Warnings);
				dim_order = new int[]{0,1,2}; 		// Default ordering
				}
			
			in_dims = dims; // new int[]{dims[dim_order[0]],dims[dim_order[1]],dims[dim_order[2]]};
			int n_voxels = dims[0]*dims[1]*dims[2];
			
			int data_type_out = DataTypes.getDataBufferType(data_type);
			
			double v_in = Double.NaN;
			
			long[] start_dims = image_ds.getStartDims();
			long[] selected_dims = image_ds.getSelectedDims();
			long[] stride = image_ds.getStride();
			selected_dims[0] = 1;
			selected_dims[1] = in_dims[1];
			selected_dims[2] = in_dims[2];
			stride[0] = 1;
			stride[1] = 1;
			stride[2] = 1;
			Object data;
			
			if (progress != null){
				progress.setMinimum(0);
				progress.setMaximum(in_dims[0]);
				}
			
			if (progress != null){
				progress.setMinimum(0);
				progress.setMaximum(in_dims[0] * 2);
				}
			ArrayList<MguiNumber> values = new ArrayList<MguiNumber>(n_voxels);
			int chunk = in_dims[1] * in_dims[2];
			int j = 0, itr = 0;
			// Fill data array
			for (int i = 0; i < n_voxels; i++){
				switch (data_type_out){
					case DataBuffer.TYPE_FLOAT:
						values.add(new MguiFloat(v_in));
						break;
					case DataBuffer.TYPE_DOUBLE:
						values.add(new MguiDouble(v_in));
						break;
					case DataBuffer.TYPE_INT:
						values.add(new MguiInteger(v_in));
						break;
					case DataBuffer.TYPE_SHORT:
					case DataBuffer.TYPE_USHORT:
					case DataBuffer.TYPE_BYTE:
						values.add(new MguiShort(v_in));
					}
				if (progress != null && ++j == chunk){
					j = 0;
					progress.update(itr++);
					}
				}
			
			for (int i = 0; i < in_dims[0]; i++){
				// Each i is a new image slice, must scale it independently
				double min, max;
				if (options.scale_slicewise && image_min != null){
					min = image_min[i];
					max = image_max[i];
				}else{
					min = range[0];
					max = range[1];
					}
				start_dims[0] = i;
				image_ds.clearData();
				data = image_ds.getData();
				min += 0;
				
				for (j = 0; j < in_dims[1]; j++){
					for (int k = 0; k < in_dims[2]; k++){
						int idx = j*in_dims[2] + k;
						switch(mgui_datatype){
							case DataBuffer.TYPE_FLOAT:
								v_in = ((float[])data)[idx];
								break;
							case DataBuffer.TYPE_DOUBLE:
								v_in = ((double[])data)[idx];
								break;
							case DataBuffer.TYPE_INT:
								v_in = ((int[])data)[idx];
								break;
							case DataBuffer.TYPE_SHORT:
							case DataBuffer.TYPE_USHORT:
								v_in = ((short[])data)[idx];
								break;
							case DataBuffer.TYPE_BYTE:
								v_in = ((byte[])data)[idx];
								break;
							default:
								v_in = ((double[])data)[idx]; // Will throw an exception
							}
						// Scale value
						v_in = scaleData(v_in, range[0], range[1], min, max);
						int abs_idx = getAbsoluteIndex(dim_order,new int[]{i,j,k});
						values.get(abs_idx).setValue(v_in);
						
						}
					}
				if (progress != null){
					progress.update(itr++);
					}
				}
			
			return values;
		
		}catch (Exception ex){
			throw new IOException(ex);
			}
		
	}

}